<?php

/**
 * Client Actions Controller
 *
 * This file contains the Actions class for handling client-related operations.
 *
 * @package ServiceEZ
 * @subpackage Controllers\Clients
 */

namespace ServiceEZ\Controllers\Clients;

use ServiceEZ\Models\Client;
use ServiceEZ\Models\Order;
use ServiceEZ\Models\Service;
use WP_REST_Request;
use WP_REST_Response;

/**
 * Actions class for handling client-related operations.
 */
class Actions {
	/**
	 * Create a new client.
	 *
	 * @param WP_REST_Request $request The REST request object.
	 * @return WP_REST_Response The REST response object.
	 */
	public function create( WP_REST_Request $request ) {
		$params = array(
			'user_id'       => get_current_user_id(),
			'firstName'     => sanitize_text_field( $request->get_param( 'firstName' ) ),
			'lastName'      => sanitize_text_field( $request->get_param( 'lastName' ) ),
			'email'         => sanitize_email( $request->get_param( 'email' ) ),
			'phone'         => sanitize_text_field( $request->get_param( 'phone' ) ),
			'address_line1' => sanitize_text_field( $request->get_param( 'address_line1' ) ),
			'address_line2' => sanitize_text_field( $request->get_param( 'address_line2' ) ),
			'city'          => sanitize_text_field( $request->get_param( 'city' ) ),
			'state'         => sanitize_text_field( $request->get_param( 'state' ) ),
			'postal_code'   => sanitize_text_field( $request->get_param( 'postal_code' ) ),
			'country'       => sanitize_text_field( $request->get_param( 'country' ) ),
		);

		// Ensure all fields are present before creating
		foreach ( $params as $key => $value ) {
			if ( $value === null ) {
				$params[ $key ] = ''; // Set empty string for null values
			}
		}

		$client = Client::updateOrCreate(
			array( 'email' => $params['email'] ),
			$params
		);
		return new WP_REST_Response( $client, 201 );
	}

	/**
	 * Get all clients.
	 *
	 * @return WP_REST_Response The REST response object containing all clients.
	 */
	public function get_all() {
		$clients = Client::all();
		return new WP_REST_Response( $clients );
	}

	/**
	 * Get a single client by ID.
	 *
	 * @param WP_REST_Request $request The REST request object.
	 * @return WP_REST_Response The REST response object containing the client or an error message.
	 */
	public function get_single( WP_REST_Request $request ) {
		$id     = $request->get_param( 'id' );
		$client = Client::find( $id );
		if ( ! $client ) {
			return new WP_REST_Response( array( 'message' => 'Client not found' ), 404 );
		}
		return new WP_REST_Response( $client );
	}

	/**
	 * Update a client.
	 *
	 * @param WP_REST_Request $request The REST request object.
	 * @return WP_REST_Response The REST response object.
	 */
	public function update( WP_REST_Request $request ) {
		$id = absint( $request->get_param( 'id' ) );

		$params = array(
			'firstName'     => sanitize_text_field( $request->get_param( 'firstName' ) ),
			'lastName'      => sanitize_text_field( $request->get_param( 'lastName' ) ),
			'email'         => sanitize_email( $request->get_param( 'email' ) ),
			'phone'         => sanitize_text_field( $request->get_param( 'phone' ) ),
			'address_line1' => sanitize_text_field( $request->get_param( 'address_line1' ) ),
			'address_line2' => sanitize_text_field( $request->get_param( 'address_line2' ) ),
			'city'          => sanitize_text_field( $request->get_param( 'city' ) ),
			'state'         => sanitize_text_field( $request->get_param( 'state' ) ),
			'postal_code'   => sanitize_text_field( $request->get_param( 'postal_code' ) ),
			'country'       => sanitize_text_field( $request->get_param( 'country' ) ),
		);

		// Remove null or undefined values.
		$params = array_filter(
			$params,
			function ( $value ) {
				return $value !== null && $value !== '';
			}
		);

		$client = Client::find( $id );
		if ( ! $client ) {
			return new WP_REST_Response( array( 'message' => 'Client not found' ), 404 );
		}

		try {
			$client->fill( $params );
			$client->save();

			return new WP_REST_Response( $client );
		} catch ( \Exception $e ) {
			return new WP_REST_Response(
				array( 'message' => 'Failed to update client: ' . $e->getMessage() ),
				500
			);
		}
	}

	/**
	 * Delete a client.
	 *
	 * @param int $id The client ID.
	 * @return WP_REST_Response The REST response object containing a success message or an error message.
	 */
	public function delete( WP_REST_Request $request ) {
		$id     = $request->get_param( 'id' );
		$client = Client::where( 'id', $id )->first();
		if ( ! $client ) {
			return new WP_REST_Response( array( 'message' => 'Client not found' ), 404 );
		}
		$client->delete();
		return new WP_REST_Response( array( 'message' => 'Client deleted successfully' ) );
	}

	public function get_client_orders( WP_REST_Request $request ) {
		$id = $request->get_param( 'id' );

		$orders = Order::where( 'client_id', $id )->get();

		return new WP_REST_Response( $orders );
	}

	/**
	 * Get client dashboard data.
	 *
	 * @param WP_REST_Request $request The REST request object.
	 * @return WP_REST_Response The REST response object containing the dashboard data.
	 */
	public function get_dashboard_data( WP_REST_Request $request ) {

		$user_id = $request->get_header( 'X-User-ID' );

		$client = Client::where( 'user_id', $user_id )->first();

		if ( ! $client ) {
			return new WP_REST_Response( array( 'message' => 'Client not found' ), 404 );
		}

		$total_orders  = Order::where( 'client_id', $client->id )->count();
		$total_spent   = Order::where( 'client_id', $client->id )->sum( 'total_amount' );
		$recent_orders = Order::where( 'client_id', $client->id )
			->orderBy( 'created_at', 'desc' )
			->limit( 5 )
			->get();

		$dashboard_data = array(
			'name'          => $client->name,
			'email'         => $client->email,
			'total_orders'  => $total_orders,
			'total_spent'   => $total_spent,
			'recent_orders' => $recent_orders,
		);

		return new WP_REST_Response( $dashboard_data );
	}
}
