<?php
/**
 * Dashboard Actions
 *
 * This file contains the actions for the dashboard.
 *
 * @package ServiceEZ\Controllers\Dashboard
 */

namespace ServiceEZ\Controllers\Dashboard;

use ServiceEZ\Models\Order;
use ServiceEZ\Models\Service;
use ServiceEZ\Models\Client;

/**
 * Class Actions
 *
 * This class contains methods for handling actions related to the dashboard.
 */
class Actions {
	/**
	 * Get the dashboard data.
	 *
	 * @return array
	 */
	public function get_dashboard_data() {
		// Fetch real data from the database.
		$total_services = $this->get_total_services();
		$total_revenue  = $this->get_total_revenue();

		$revenue_over_time = $this->get_revenue_over_time();
		$services_by_type  = $this->get_services_by_type();

		$total_orders  = $this->get_total_orders();
		$total_clients = $this->get_total_clients();

		$data = array(
			'total_services'    => $total_services,
			'total_revenue'     => $total_revenue,
			'revenue_over_time' => $revenue_over_time,
			'services_by_type'  => $services_by_type,
			'total_orders'      => $total_orders,
			'total_clients'     => $total_clients,
		);

		return $data;
	}

	/**
	 * Get the total number of services.
	 *
	 * @return int
	 */
	private function get_total_services() {
		return Service::count();
	}

	/**
	 * Get the total number of orders.
	 *
	 * @return int
	 */
	private function get_total_orders() {
		return Order::count();
	}

	/**
	 * Get the total number of clients.
	 *
	 * @return int
	 */
	private function get_total_clients() {
		return Client::count();
	}

	/**
	 * Get the total revenue.
	 *
	 * @return int
	 */
	private function get_total_revenue() {
		return (int) Order::where( 'order_status', 'active' )->orWhere( 'order_status', 'completed' )->sum( 'total_amount' );
	}


	/**
	 * Get the revenue over time.
	 *
	 * @return array
	 */
	private function get_revenue_over_time() {
		$revenue_data = Order::where( 'order_status', 'active' )
			->orWhere( 'order_status', 'completed' )
			->get();

		$revenue_data = $revenue_data->map(
			function ( $item ) {
				return array(
					'date'    => $item->created_at,
					'revenue' => $item->total_amount,
				);
			}
		);
		return $revenue_data;
	}

	/**
	 * Get the number of services by type.
	 *
	 * @return array
	 */
	private function get_services_by_type() {
		// Group services by type and count them.
		$services         = Service::all();
		$services_by_type = $services->groupBy( 'service_type' )->map(
			function ( $service ) {
				return $service->count();
			}
		);

		return $services_by_type;
	}
}
