<?php
/**
 * Notifications controller.
 *
 * @package ServiceEZ\Controllers\Notifications
 */

namespace ServiceEZ\Controllers\Notifications;

use ServiceEZ\Models\Notification;
use WP_REST_Request;
use WP_REST_Response;

class Actions {
	/**
	 * Get all notifications.
	 *
	 * @param WP_REST_Request $request The request object.
	 * @return WP_REST_Response The response object.
	 */
	public function get_all( WP_REST_Request $request ) {
		$user_id = get_current_user_id();

		$page     = $request->get_param( 'page' ) ?? 1;
		$per_page = $request->get_param( 'per_page' ) ?? 20;

		// If user is not admin, return only notifications for the current user.
		if ( ! current_user_can( 'administrator' ) ) {
			$user_id       = $request->get_header( 'X-User-Id' );
			$notifications = Notification::with( 'user' )
				->where( 'user_id', $user_id )
				->orderBy( 'created_at', 'desc' )
				->skip( ( $page - 1 ) * $per_page )
				->take( $per_page )
				->get()
				->map(
					function ( $notification ) {
						$notification->created_at = get_date_from_gmt( $notification->created_at );
						return $notification;
					}
				);

			return new WP_REST_Response( $notifications );
		}

		$notifications = Notification::with( 'user' )
			->orderBy( 'created_at', 'desc' )
			->skip( ( $page - 1 ) * $per_page )
			->take( $per_page )
			->get()
			->map(
				function ( $notification ) {
					$notification->created_at = get_date_from_gmt( $notification->created_at );
					return $notification;
				}
			);

		return new WP_REST_Response( $notifications );
	}

	/**
	 * Delete a notification.
	 *
	 * @param WP_REST_Request $request The request object.
	 * @return WP_REST_Response The response object.
	 */
	public function delete( WP_REST_Request $request ) {
		$id = $request->get_param( 'id' );

		if ( ! is_numeric( $id ) ) {
			return new WP_REST_Response( array( 'message' => 'Invalid notification ID' ), 400 );
		}

		$notification = Notification::find( (int) $id );
		if ( ! $notification ) {
			return new WP_REST_Response( array( 'message' => 'Notification not found' ), 404 );
		}

		$notification->delete();
		return new WP_REST_Response( array( 'message' => 'Notification deleted successfully' ) );
	}

	/**
	 * Delete all notifications for a user.
	 *
	 * @param WP_REST_Request $request The request object.
	 * @return WP_REST_Response The response object.
	 */
	public function delete_all( WP_REST_Request $request ) {
		$user_id = get_current_user_id();
		Notification::where( 'user_id', $user_id )->delete();
		return new WP_REST_Response( array( 'message' => 'All notifications deleted successfully' ) );
	}

	/**
	 * Mark a notification as read.
	 *
	 * @param WP_REST_Request $request The request object.
	 * @return WP_REST_Response The response object.
	 */
	public function mark_as_read( WP_REST_Request $request ) {
		$id = $request->get_param( 'id' );

		if ( ! is_numeric( $id ) ) {
			return new WP_REST_Response( array( 'message' => 'Invalid notification ID' ), 400 );
		}

		$notification = Notification::find( (int) $id );
		if ( ! $notification ) {
			return new WP_REST_Response( array( 'message' => 'Notification not found' ), 404 );
		}

		$notification->read_at = current_time( 'mysql', true );
		$notification->save();

		return new WP_REST_Response( array( 'message' => 'Notification marked as read' ) );
	}
}
