<?php
/**
 * Frontend class.
 *
 * @package ServiceEZ
 * @subpackage Frontend
 * @since 1.0.0
 */

declare(strict_types=1);

namespace ServiceEZ\Frontend;

use ServiceEZ\Core\Template;
use ServiceEZ\Traits\Base;
use ServiceEZ\Libs\Assets;
/**
 * Class Frontend
 *
 * Handles frontend functionalities for the ServiceEZ.
 *
 * @package ServiceEZ\Frontend
 */
class Frontend {

	use Base;

	/**
	 * Script handle for ServiceEZ.
	 */
	const HANDLE = 'serviceez';

	/**
	 * JS Object name for ServiceEZ.
	 */
	const OBJ_NAME = 'serviceez';

	/**
	 * Development script path for ServiceEZ.
	 */
	const DEV_SCRIPT = 'src/admin/main.jsx';

	/**
	 * List of allowed screens for script enqueue.
	 *
	 * @var array
	 */
	private $allowed_screens = array(
		'toplevel_page_serviceez',
	);

	/**
	 * Frontend bootstrapper.
	 *
	 * @return void
	 */
	public function bootstrap() {
		add_action( 'wp_enqueue_scripts', array( $this, 'enqueue_script' ) );
		add_action( 'admin_enqueue_scripts', array( $this, 'enqueue_script' ) );
	}

	/**
	 * Enqueue script based on the current screen.
	 *
	 * @param string $screen The current screen.
	 */
	public function enqueue_script( $screen ) {
		$current_screen     = $screen;
		$template_file_name = Template::FRONTEND_TEMPLATE;

		if ( ! is_admin() ) {
			$template_slug = get_page_template_slug();
			if ( $template_slug ) {

				if ( $template_slug === $template_file_name ) {
					array_push( $this->allowed_screens, $template_file_name );
					$current_screen = $template_file_name;
				}
			}
		}

		if ( in_array( $current_screen, $this->allowed_screens, true ) ) {
			wp_enqueue_media();
			Assets\enqueue_asset(
				SEZ_DIR . 'assets/admin/dist',
				self::DEV_SCRIPT,
				$this->get_config()
			);
			wp_localize_script( self::HANDLE, self::OBJ_NAME, $this->get_data() );

			// Remove all admin notices.
			remove_all_actions( 'admin_notices' );
		}
	}

	/**
	 * Get the script configuration.
	 *
	 * @return array The script configuration.
	 */
	public function get_config() {
		return array(
			'dependencies' => array( 'react' ),
			'handle'       => self::HANDLE,
			'in-footer'    => true,
		);
	}

	/**
	 * Get data for script localization.
	 *
	 * @return array The localized script data.
	 */
	public function get_data() {

		return array(
			'isAdmin'      => is_admin(),
			'siteUrl'      => esc_url( home_url() ),
			'adminUrl'     => esc_url( admin_url() ),
			'apiUrl'       => esc_url_raw( rest_url() ),
			'userInfo'     => $this->get_user_data(),
			'templateUrl'  => esc_url( Template::get_template_slug_used_on_page()[0]->guid ),
			'assetsUrl'    => esc_url( SEZ_ASSETS_URL ),
			'isLoggedIn'   => is_user_logged_in(),
			'logoutUrl'    => esc_url( wp_logout_url() ),
			'nonce'        => wp_create_nonce( 'wp_rest' ),
			'userId'       => absint( get_current_user_id() ),
			'userRole'     => esc_attr( $this->get_user_role() ),
			'defaults'     => array(
				'logo' => SEZ_ASSETS_URL . '/logo.svg',
			),
			'isOnboarding' => get_option( 'serviceez_onboarding', false ),
			'language'     => isset( get_option( 'serviceez_settings' )['general']['language'] ) ? get_option( 'serviceez_settings' )['general']['language'] : '',
		);
	}


	/**
	 * Get the user role.
	 *
	 * @return string The user role.
	 */
	public function get_user_role() {
		if ( is_user_logged_in() ) {
			$current_user = wp_get_current_user();
			return count( $current_user->roles ) > 0 ? $current_user->roles[0] : '';
		}
		return '';
	}


	/**
	 * Get user data for script localization.
	 *
	 * @return array The user data.
	 */
	private function get_user_data() {
		$username   = '';
		$avatar_url = '';
		$email      = '';

		if ( is_user_logged_in() ) {
			// Get current user's data .
			$current_user = wp_get_current_user();

			// Get username.
			$username = $current_user->user_login; // or use user_nicename, display_name, etc.

			// Get avatar URL.
			$avatar_url = get_avatar_url( $current_user->ID );

			// Get email.
			$email = $current_user->user_email;
		}

		return array(
			'username' => $username,
			'avatar'   => $avatar_url,
			'email'    => $email,
		);
	}
}
