<?php
use ServiceEZ\Core\Api;
use ServiceEZ\Admin\Menu;
use ServiceEZ\Core\Template;
use ServiceEZ\Frontend\Frontend;
use ServiceEZ\Traits\Base;

defined( 'ABSPATH' ) || exit;

/**
 * Class ServiceEZ
 *
 * The main class for the ServiceEZ plugin, responsible for initialization and setup.
 *
 * @since 1.0.0
 */
final class ServiceEZ {

	use Base;

	/**
	 * License controller instance
	 *
	 * @var LicenseActions
	 */
	private $license_controller;

	/**
	 * Class constructor to set up constants for the plugin.
	 *
	 * @since 1.0.0
	 * @return void
	 */
	public function __construct() {
		define( 'SEZ_VERSION', '1.0.0' );
		define( 'SEZ_PLUGIN_FILE', __FILE__ );
		define( 'SEZ_DIR', plugin_dir_path( __FILE__ ) );
		define( 'SEZ_URL', plugin_dir_url( __FILE__ ) );
		define( 'SEZ_ASSETS_URL', SEZ_URL . 'assets' );
		define( 'SEZ_ROUTE_PREFIX', 'serviceez/v1' );

	}

	/**
	 * Main execution point where the plugin will fire up.
	 *
	 * Initializes necessary components for both admin and frontend.
	 *
	 * @since 1.0.0
	 * @return void
	 */
	public function init() {
		if ( is_admin() ) {
			Menu::get_instance()->init();
		}

		// Initialize core functionalities
		Frontend::get_instance()->bootstrap();
		API::get_instance()->init();
		Template::get_instance()->init();

		// Set up license check schedule
		add_action( 'init', array( $this, 'setup_license_check_schedule' ) );
		add_action( 'serviceez_check_license', array( $this->license_controller, 'check_license_status' ) );

		add_action( 'init', array( $this, 'i18n' ) );
	}


	/**
	 * Sets up the daily license check schedule.
	 *
	 * @return void
	 */
	public function setup_license_check_schedule() {
		if ( ! wp_next_scheduled( 'serviceez_check_license' ) ) {
			wp_schedule_event( time(), 'daily', 'serviceez_check_license' );
		}
	}

	/**
	 * Internationalization setup for language translations.
	 *
	 * Loads the plugin text domain for localization.
	 *
	 * @since 1.0.0
	 * @return void
	 */
	public function i18n() {
		load_plugin_textdomain( 'serviceez', false, dirname( plugin_basename( __FILE__ ) ) . '/languages/' );
	}

	/**
	 * Plugin deactivation hook callback.
	 * Cleans up scheduled events.
	 *
	 * @return void
	 */
	public static function deactivate() {
		wp_clear_scheduled_hook( 'serviceez_check_license' );
	}
}

// Register deactivation hook
register_deactivation_hook( __FILE__, array( 'ServiceEZ', 'deactivate' ) );
